/* ============================================================================================

  The aim is to simulate the power of the test statistics of Genest et al 2006 (Revision)
  For AR(1) residual.
  The true model is X_t=\mu + \phi X_{t-1} + \ve_t$ where $\ve_t=  \alpha ve_{t-1} + u_t$ with
  $u_t  N(0,1)$ and   $\alpha = 0, 0.2, 0.4, 0.6 0.8$
  $\phi=0.3$ and $mu=-2$.  
  
    We shall consider all the statistics outlined in the paper.
===============================================================================================*/


#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <malloc.h>
#include <sys/types.h>
#include <time.h>

int Bias_Correction;
/* #include "kilani/kiss.c" */
#include "qsort.c"
#include "bds_prod_integral_comp.c"
#include "bds_g.c"
#include "bds_stat_lebaron.c"
#include "series.c"
#include "rang.c"

/**********************************************************************/ 
/* programme:                                                         */ 
/**********************************************************************/ 
main()  
{  
    
    FILE *fp1, *fp2;
    double eps, stat[20],  *u, *x, *e, *uu, reject[10][5],seuil[10][5];
    double **MM, *NN, cte[10],var[10];
    long int it, i, j, k, isim,ialpha; 
    int n, p, *rangs;
    long int *ii;   
    unsigned long int i0, j00, k0;   
    double mu, phi, alpha,muhat,phihat,sum1,sum2,sum3;
    time_t *tloc;   
    static char *stat_name[9] = {"$S_n$","$\\tilde S_n^\\star$", "$\\tilde I_n^\\star$", "$\\tilde M_n^\\star$","$\\tilde T_n^\\star$",
        "$\\tilde S_n$","$\\tilde I_n$","$\\tilde M_n$","$\\tilde T_n$"};
    
    
    /****  Fixation du seed  ****/ 
    ii = (long int *)malloc(sizeof(long int));   
    tloc=(time_t*)malloc(sizeof(time_t));   
    *ii=time(tloc);   
    j00=*ii*rand()*123;   
    k0=*ii*rand()*756;   
    i0=*ii*rand()*6675;   
    kisset(i0,j00,k0);     
    
    
    p =6;
    
    
    it = 10000;   /** Nombre d'iterations **/ 
    n=100; /* sample size */
    
    
    x = (double *)calloc((n+30),sizeof(double)); 
    e = (double *)calloc((n+30),sizeof(double)); 
    u = (double *)calloc(n,sizeof(double)); 
    uu = (double *)calloc(n,sizeof(double)); 
    rangs = (int *)calloc(n+p,sizeof(int));
    NN = calloc(n+1,sizeof(double)); 
    MM = calloc(n+1,sizeof(double)); 
    for(i=0; i<=n; i++){MM[i] = calloc(n+1,sizeof(double)); }
    
    /*=============================================================
    Critical Values from JASA Paper      
    For each statisitic the critical values are given for
    each valu of m considered 2...6
    The order of the statistics is:
    Row 0: BDS,
    Row 1: $\tilde S_n^*$,
    Row 2: $\tilde I_n^*$,
    Row 3: $\tilde M_n^*$,
    Row 4: $\tilde T_n^*$,
    Row 5: $\tilde S_n$,
    Row 6: $\tilde I_n$,
    Row 7: $\tilde M_n$,
    Row 8: $\tilde T_n$,
    ==============================================================*/
    fp1=fopen("seuil.txt","r");
    
    for(i=0;i<9;i++)
    {
        for(j=2;j<=p;j++)
        {
            fscanf(fp1,"%lf ",&seuil[i][j-2]);
            
        }
    }
    fclose(fp1);
    
    fp2=fopen("Power_residual_res.text","a");
      fprintf(fp2, "\\begin{center}\n{\\sc Tableau }: Percentage of rejection  \\\\");  
      fprintf(fp2, "\\begin{tabular}{llllll} \n\\hline \n"); 
      fprintf(fp2," $alpha$ & Stat & $p=2$ &  $p=3$ &  $p=4$ & $p=5$ & $p=6$ \\\\\\hline \n\n");     

    for(ialpha=0; ialpha<5;ialpha++)
    {
        alpha=1.0*ialpha/5;
        for(i=0;i<9;i++)
        {
            for(j=2;j<=p;j++)
            {
                reject[i][j-2]=0.0;
            }
        }
        for(isim=1;isim<=it;isim++)
        {
            /*=============================================================*/
            /* Generate the data from the model above                      */
            /*=============================================================*/
            mu=-2.0; phi=0.3;
            e[0]=rnorm();
            x[0]=mu+e[0];
            for(j=1;j<n+30;j++)
            {
                e[j]=alpha*e[j-1]+rnorm();
                x[j]=mu+phi*x[j-1]+ e[j];
            }
            for(j=0;j<n;j++) u[j]=x[j+30];
            /* The vector u contains the series coming for the above model ARMA(1,1). */
            
            /*==============================================*/
            /* Fitting and AR(1) to the series is done next */
            /*==============================================*/
            sum1=0.0;
            sum2=0.0;
            sum3=0.0;
            for(j=0;j<n-1;j++) 
            {
                sum1 +=u[j];
                sum3 +=u[j]*u[j];
            }
            sum1=sum1; // X bar
            for(j=0;j<n-1;j++) 
            {
                
                sum2 +=u[j]*u[j+1];
                
            }
            
            phihat=(sum2-(sum1+u[n-1]-u[0])*sum1/(n-1))/(sum3-sum1*sum1/(n-1));
            muhat=sum1*(1.0-phihat)/(n-1);
            /* printf("\n %f  %f",muhat,phihat); */
            
            /*   computing the residuals uu*/
            /* muhat=mu; phihat=phi;*/ /*  just to check  the levels */
            uu[0]=u[0]-muhat;
            for(j=1;j<n;j++)
            {
                uu[j]=u[j]-muhat-phihat*u[j-1];
            }
            rank(uu, n, rangs);
            
            for(j=n;j<n+p;j++)
                rangs[j] = rangs[j-n]; /* for the circular definition */
            
            /* Calling the statistical procedures */
            
            /* classical BDS */
            eps=1.0;
            bdsstat(uu, n, p, eps, stat);
            for(k=2;k<=p;k++)
            {
                if (fabs(stat[k])>seuil[0][k-2]) reject[0][k-2]+=1.0;
            }
            
  
            if( (isim % 100) == 0) printf("%d ", isim/100); 
            
            
            
      }     
      for(i=0;i<1;i++)
      {
          fprintf(fp2,"\n %f & %20s  ", alpha, stat_name[i]);
          for(j=2;j<=p;j++)
          {
              fprintf(fp2," & %8.2f ",reject[i][j-2]/it*100);
          }
          fprintf(fp2,"\\\\");
      }
      fprintf(fp2,"\\hline");
      

    }
    fprintf(fp2,"\n \n n= %d, \n The critical value used are ",n);
      for(i=0;i<1;i++)
      {
          fprintf(fp2,"\n %20s  ", stat_name[i]);
          for(j=2;j<=p;j++)
          {
              fprintf(fp2,"%f ",seuil[i][j-2]);
          }
      }
    free(u) ; free(rangs); free(x), free(e); free(uu);
    
    
} 
