/************************************************************************/
/****  Date: April 29, 2003,    Bruno Rmillard                       ***/
/****  Email:  bruno.remillard@hec.ca                                 ***/
/****  Calculate the Cramer-von Mises statistics of                   ***/
/****  the BDS                                                        ***/
/****  Commentd added Sep 16, 2006                                    ***/  
/************************************************************************/



/* #include "../../LibC/rang.c" */

double mini( double u, double v)
{
    if(u < v)
        return u;
    else
        return v;
}

double maxi( double u, double v)
{
    if(u < v)
        return v;
    else
        return u;
}

double Gtilde_n(int m, int n,int correct)
{
	double t0,t1,t2,t3,t4,t7,t16,t19,t20,t22,t23,t27,t28,t29;
	if (correct==0) 
	{
		t0=1.0/(m+1.0);
	}
	else
	{
		switch(m)
		{
		case 2:
			t1 = n*n;
			t0 = (5.0*t1*n+t1-4.0)/n/(t1-1.0)/15.0;
			break;
		case 3:
			t1 = n*n;
			t2 = t1*t1;
			t16 = pow(n-1.0,2.0);
			t0 = (105.0*t2*n-36.0*t2-49.0*t1*n-120.0*t1+196.0*n-64.0)/t1/(n+1.0)/t16
				/420.0;
			break;
		case 4:
			t1 = n*n;
			t2 = t1*n;
			t3 = t1*t1;
			t19 = n-1.0;
			t20 = t19*t19;
			t0 = (63.0*t3*t2-59.0*t3*t1-39.0*t3*n-65.0*t3+192.0*t2-44.0*t1+24.0*n
				-48.0)/t2/(n+1.0)/t20/t19/315.0;
			break;
		case 5:
			t1 = n*n;      
			t2 = t1*t1;
			t3 = t2*t2;
			t7 = t1*n;
			t22 = pow(n-1.0,2.0);
			t23 = t22*t22;
			t0 = (231.0*t3*n-362.0*t3-88.0*t2*t7-142.0*t2*t1+1199.0*t2*n-736.0*t2
				-1342.0*t7+2608.0*t1-792.0*n-384.0)/t2/t23/(n+1.0)/1386.0;
			
			break;
		case 6:
			t1 = n*n;
			t2 = t1*n;
			t3 = t1*t1;
			t4 = t3*t3;
			t16 = t3*n;
			t22 = 2145.0*t4*t2-4775.0*t4*t1+650.0*t4*n-290.0*t4+15951.0*t3*t2-18165.0
				*t3*t1-14794.0*t16+51190.0*t3-94016.0*t2+128696.0*t1-48256.0*n-12224.0;
			t27 = n-1.0;
			t28 = t27*t27;
			t29 = t28*t28;
			t0 = t22/t16/(n+1.0)/t29/t27/15015.0;
			break;
		}
	}
	return t0;
}

double M(double u, double v)   /*** M(u,v) with dG ***/

{
    double u1,v1,s,t, s3, t3;
    u1 = mini(u,1.0-u);
    v1 = mini(v,1.0-v); 
    s = mini(u1,v1); 
    t = maxi(u1,v1);
    t3 = t*t*t;
    s3 = s*s*s;
    
    
    return  1.0/6.0 - s3 -2.0*t3/3.0 + s*(1.0+s3)/3.0    +t*(1.0+t3)/3.0    +s*t*(1.0-t);
}


double N(double u)   /*** N(u) with dG ***/

{
    double v;
    
    v  = (u-0.5)*(u-0.5);
    
    return 187.0/480.0 -0.75*v + 0.5*v*v;
}



void matM(int n, double *NN, double **MM)
{
    int i,j,k;
    double t,t0,v,t1,t2,t3,t6,t20,t17,t30,t43,t45,t46;
    NN[0] = 1.0;
    MM[0][0] = 1.0;
    for(i=1;i<=n;i++)
    {
        MM[i][0] = 1.0;
        MM[0][i] = 1.0;
        t =  1.0*i/((double) n+1);
		v=(t-0.5)*(t-0.5);
		if(Bias_Correction==0) t0=187.0/480.0 -0.75*v + 0.5*v*v;
		k=min(i,n+1-i);
		if(Bias_Correction==1) 
		{
			t0 = (-60.0*k*n+30.0*k*k*k-15.0*k*k-15.0*k*k*k*k+45.0*k*k*k*k*n-60.0*k*k*k*n+75.0*n*k*k-90.0*k*k*k*n*n+45.0*k*n*n*n*n+90.0*n*n*k*k+30.0*k*n*n*n-75.0*k*n
				*n-26.0*n+21.0*n*n*n*n*n-45.0*n*n+45.0*n*n*n*n+5.0*n*n*n)/n/pow(n+1.0,3.0)/(n
				-1.0)/90.0;
			
		}
		
        NN[i] = t0;
        MM[i][i] = M(t,t);
        for(j=1;j<i;j++)
        {
            MM[i][j] = M(t, 1.0*j/((double) n+1));
            MM[j][i] = MM[i][j];
        }
    }
}



void newmatM(int n, double *NN, double **MM)

{
    int i,j,k;
    double t, t2;
    for(i=0;i<=n;i++)
    {
        t = 1.0-  1.0*i/((double) n+1);
        t2 = t*t;
		k=i;
        if(Bias_Correction==0) NN[i] = t2-0.5*t2*t2;
		if(Bias_Correction==1) NN[i]=(12.0*k*k*k*n-12.0*n*n*k*k+12.0*n*n*k+6.0*n*k-18.0*n*k*k-2.0*n-4.0*k-3.0*n*n+3.0*k*k+4.0*k*k*k-3.0*k*k*k*k+3.0*n*n*n*n+2.0*n*n*n)/n/(n-1.0)/pow(n+
			1.0,2.0)/6.0;
		
        MM[i][i] = t2;
        for(j=0;j<i;j++)
        {
            MM[i][j] = t2;
            MM[j][i] = MM[i][j];
        }
    }
}


void matM_BDS(int n, double **MM, double eps)/* Setting MM for tilde Sn */

{
    int i,j;
    double t;
    
    t =  eps*(n+1);
    for(i=1;i<=n;i++)
    {
        MM[i][i] = 1.0;
        for(j=1;j<i;j++)
        {
            MM[i][j] = (i-j <= t);
            MM[j][i] = MM[i][j];
        }
    }
}


double tilde_F(double u, double t)
{
    double s = u+t;
    double v = u-t;
    
    
    if(s< 1.0) { 
        if( v>0.0) 
            return s-v; 
        else 
            return s;}
    else { 
        if( v>0.0) 
            return 1.0-v; 
        else 
            return 1.0;}
}


void vec_BDS(int n, double *NN, double eps)

{
    int i;
    double t;
    
    
    for(i=1;i<=n;i++)
    {
        t =  1.0*i/((double) n+1);
        NN[i] = tilde_F(t,eps);
    }  
}


void puiss(int p, int n, double *c)
{
    int k,m;
    double t;
    if(Bias_Correction==0) t = 1.0/3.0;
	if(Bias_Correction==1) 
	{
		t=(5.0*n*n*n+n*n-4.0)/n/(n*n-1.0)/15.0;
	}
	c[0]=1.0;
    for(k=1;k<=p;k++){ c[k] = t*c[k-1]; }
}


void puiss_BDS(int p, int n, double *c, double eps)

{
    int k,m;
    double t;
    if(Bias_Correction==0) t = 2*eps-eps*eps;
	if(Bias_Correction==1) 
	{
		m=floor((n+1)*eps);
		t=2*m/(n-1.0)-m*(m+1.0)/(n*(n-1.0));
	}
    
    c[0] = 1.0;
    
    for(k=1;k<=p;k++){ c[k] = t*c[k-1]; }
}
void puiss_BDS_star(int p, int n, double *c, double eps)

{
    int k,m;
    double t,t0;
    if(Bias_Correction==0) t0 = 2*eps-eps*eps;
	if(Bias_Correction==1) 
	{
		t=eps;
		k=floor((n+1)*min(t,1-t));
		if(t<= 0.5) t0 = (-2.0*t*k*n+t+k*k-2.0*t*k+3.0*t*n+2.0*t*n*n)/(n+1.0)/n;
        if(t>0.5) t0 = (1.0+2.0*t*k*n+2.0*n-t-2.0*k+k*k-2.0*k*n+2.0*t*k-t*n+n*n)/(n+1.0)/n;
	}
    t=t0;
    c[0] = 1.0;
    
    for(k=1;k<=p;k++){ c[k] = t*c[k-1]; }
}

void bds_rank(int *rangs, int n, int p, double *stat, double **MM, double *NN, double *cte)  /*** TN_star  ***/
/*  =====  The statistics $\tilde T_n^\star$ in our JASA 2006 Paper  ====*/

/* Before calling this routine you need  to set MM, NN and cte
MM and  NN are set via the program matM
cte is set by the program puiss
*/
{
    
    int i,j,k;
    
    double m,a, b, c, *sum1, *sum2, *sum3;
    
    sum1 = (double *)calloc(p+1,sizeof(double));
    sum2 = (double *)calloc(p+1,sizeof(double));
    sum3 = (double *)calloc(p+1,sizeof(double));
    
    
    for(k=2;k<=p;k++)
    {
        sum1[k] = 0.0; 
        sum2[k] = 0.0;
        sum3[k] = 0.0;
    }
    
    
    
    for(i=0; i<n ;i++)
    {  
        b = NN[rangs[i]];
        c = MM[rangs[i]][rangs[i]]; 
        
        for(k=2;k<=p;k++)
        {
            b *= NN[rangs[i+k-1]];
            sum2[k] += b;
            c *= MM[rangs[i+k-1]][rangs[i+k-1]];
            sum3[k] += c; 
            
        }    
        for(j=0;j< i;j++)
        {
            a = MM[rangs[i]][rangs[j]];  
            
            for(k=2;k<=p;k++)
            {
                a *= MM[rangs[i+k-1]][rangs[j+k-1]];  
                sum1[k] += a;
            }
        }
    }
    
    
    for(k=2;k<=p;k++){
        m = (double) n;
        stat[k] =  m*cte[k] -2.0*sum2[k] + 2.0*sum1[k]/m +sum3[k]/m;
        stat[k]=4*stat[k]; /* to have the same definition as in the JASA Paper */
    }
    
    free(sum1); free(sum2); free(sum3);
} 


void var_bds(double eps, int p, double *var)
{
    int k, j;
    double v, g, eps2,sigma;
    
    eps2 = eps*eps;
    v= 2*eps-eps2;
    
    if(eps < 0.5)
        g = 4.0*eps2-10.0*eps*eps2/3.0;
    else
        g = -1.0/3.0 +2.0*eps-2*eps*eps2/3.0;
    
    for(k=2;k<=p;k++)
    { 
        sigma = 0.0;
        for(j=1;j<=k-1;j++) {
            sigma += 2.0*pow(g,k-j)*pow(v,2*j);
        }
        sigma += pow(g,k) + (k-1)*(k-1)*pow(v,2*k)
            -k*k*g*pow(v,2*k-2);
        sigma *= (double)4;
        var[k] = sigma;
    }
    
} 


void bds_rank_lebaron(int *rangs, int n, int p, double *stat, double **MM, double *cte, double  *var)   /*** SN  ***/
/*  =====  The statistics $\tilde S_n$ in our JASA 2006 Paper  ====*/

/* Before calling you need to set MM,  cte and var.
MM is set by the routine matM_BDS
cte is set by the routine puiss_BDS
var is set by the routine var_BDS
*/

{
    
    int i,j,k;
    
    double m, a, *sum1;
    
    
    sum1 = (double *)calloc(p+1,sizeof(double));
    
    
    
    
    for(k=2;k<=p;k++)
    {
        sum1[k] = 0.0; 
    }
    
    for(i=0; i<n ;i++)
    {     
        for(j=0;j< i;j++)
        {
            a = MM[rangs[i]][rangs[j]];  
            
            for(k=2;k<=p;k++)
            {
                a *= MM[rangs[i+k-1]][rangs[j+k-1]];  
                sum1[k] += a;
            }
        }
    }
    
    m = (double) n;
    
    for(k=2;k<=p;k++){
        
        stat[k] =  sqrt(m)*(2.0*sum1[k]/((m-1)*m) -cte[k])/sqrt(var[k]);
    }
    free(sum1);
} 



void bds_rank_lebaron_prod(int *rangs, int n, int p, double *stat, double *NN, double *cte, double  *var)  /*** SN_star  ***/
/*  =====  The statistics $\tilde S_n^\star$ in our JASA 2006 Paper  ====*/

/* Before calling you need to set NN,  cte and var.
NN is set by the routine vec_BDS
cte is set by the routine puiss_BDS
var is set by the routine var_BDS
*/

{
    
    int i,k;
    
    double m,a, *sum1;
    
    sum1 = (double *)calloc(p+1,sizeof(double));
    
    for(k=2;k<=p;k++)
    {
        sum1[k] = 0.0; 
    }
    
    for(i=0; i<n ;i++)
    {  
        a = NN[rangs[i]];  
        for(k=2;k<=p;k++)
        {
            a *= NN[rangs[i+k-1]];  
            sum1[k] += a;
        }
    }
    
    m = (double) n;
    for(k=2;k<=p;k++){
        
        stat[k] =  2.0*sqrt(m)*(sum1[k]/m -cte[k])/sqrt(var[k]);
    }
    free(sum1);
} 



void bds_rank_2(int *rangs, int n, int p, double *stat, double **MM, double *NN, double *cte)   /*** TN  ***/
/*  =====  The statistics $\tilde T_n $ in our JASA 2006 Paper  ====*/

/* Before calling you need to set MM, NN and cte.
MM and NN are set by the routine newmatM
cte is set by the routine puiss
*/

{
    
    int k, i1, i2, j1, j2, t1, t2, t3;
    
    double m,a, b, *sum1, *sum2, *sum3;
    
    sum1 = (double *)calloc(p+1,sizeof(double));
    sum2 = (double *)calloc(p+1,sizeof(double));
	sum3 = (double *)calloc(p+1,sizeof(double));
    
    
    for(k=2;k<=p;k++)
    {
        sum1[k] = 0.0; 
        sum2[k] = 0.0;
		sum3[k] = 0.0;
    }
    
    for(i1=1; i1<n ;i1++)
    {  
        
        for(j1=0;j1< i1;j1++)
        {
            t1 = abs(rangs[i1]-rangs[j1]);
            
            
            
            for(i2=1; i2<i1 ;i2++)
            {  
                for(j2=0;j2< i2;j2++)
                {
                    t2 = abs(rangs[i2]-rangs[j2]);
                    a = MM[t1][t2]; 
                    
                    for(k=1;k<p;k++)
                    {
                        t3 = abs(rangs[i1+k]-rangs[j1+k]);
                        t2 = abs(rangs[i2+k]-rangs[j2+k]);
                        a *= MM[t3][t2]; 
                        sum1[k+1] += a;
                    }
                }
            }
            
            
        }
    }
	for(i1=1; i1<n ;i1++)
    {  
        
        for(j1=0;j1< i1;j1++)
        {
            t1 = abs(rangs[i1]-rangs[j1]);
            
            
              
                for(j2=0;j2< i1;j2++)
                {
                    t2 = abs(rangs[i1]-rangs[j2]);
                    a = MM[t1][t2]; 
                    
                    for(k=1;k<p;k++)
                    {
                        t3 = abs(rangs[i1+k]-rangs[j1+k]);
                        t2 = abs(rangs[i1+k]-rangs[j2+k]);
                        a *= MM[t3][t2]; 
                        sum3[k+1] += a;
                    }
                }
           
            
            
        }
    }
    
    for(i1=1; i1<n ;i1++)
    {  
        
        for(j1=0;j1< i1;j1++)
        {
            t1 = abs(rangs[i1]-rangs[j1]);
            b  = NN[t1];
            for(k=1;k<p;k++)
            {
                t1 = abs(rangs[i1+k]-rangs[j1+k]);
                b *= NN[t1]; 
                sum2[k+1] += b;
            }
            
        }
    }
    
    
    m = (double) n;
    for(k=2;k<=p;k++){
        
        stat[k] = m*cte[k] -4.0*sum2[k]/(m-1.0) + 4.0*(2*sum1[k]+sum3[k])/(m*(m-1.0)*(m-1.0)) ;
    }
    
    free(sum1); free(sum2); free(sum3);
}


void bds_rank_int(int *rangs, int n, int p, double *stat)     /*** IN  ***/
/*  =====  The statistics $\tilde I_n in our JASA 2006 Paper  ====*/

{
    
    int i,j,k, a,b,c, *sum1;
    
    double m ;
    
    m = 1.0/((double ) n+1);
    
    sum1 = (int *)calloc(p+1,sizeof(int));
    
    for(k=2;k<=p;k++)
    {
        sum1[k] = 0; 
    }
    
    for(i=1; i<n ;i++)
    {     
        for(j = 0; j < i; j++)
        {
            a = n+1 - abs(rangs[i]-rangs[j]);
            b = a*a;
            
            /* printf("\n %d %d %d %d  %d", rangs[i], rangs[j], abs(rangs[i]- rangs[j]), b); */
            
            for(k=2;k<=p;k++)
            {
                a = n+1 - abs(rangs[i+k-1]-rangs[j+k-1]);
                c = a*a;
                if(c < b)
                    b = c;
                /* printf("\n %d %d  a = %d  b = %d c = %d" , rangs[i], rangs[j], a, b, c); */
                sum1[k] += b;
            }
        }
    }
    
    m = (double) n;
    
    for(k=2;k<=p;k++)
        stat[k] =  sqrt(m)*(  2.0*sum1[k]/((m*m-1.0)*m*(m+1.0))  -Gtilde_n(k,n,Bias_Correction)  );
    
    free(sum1);
} 




void bds_rank_lebaron_sup(int *rangs, int n, int p, double *stat)   /*** MN  ***/
/*  =====  The statistics $\tilde M_n $ in our JASA 2006 Paper  ====*/
{
    
    double **MM, *var, eps, *cte, *ks;
    int iep,j,k;
    
    cte = (double *)calloc(p+1,sizeof(double));
    var = (double *)calloc(p+1,sizeof(double));
    ks  = (double *)calloc(p+1,sizeof(double));
    
    for(k=2;k<=p;k++)
    {
        var[k] = 1.0;
        stat[k] = 0.0;
    }   
    
    
    MM = (double **)calloc(n+1,sizeof(double *)); 
    for(j=0; j<=n; j++)
        MM[j] = (double *)calloc(n+1,sizeof(double));
    
    for(iep=1;iep<=n;iep++)
    {
        eps = 1.0*iep/((double) n+1);
        puiss_BDS(p,n, cte, eps);
        matM_BDS(n, MM, eps);
        
        bds_rank_lebaron(rangs, n, p, ks, MM, cte, var);
        
        for(k=2;k<=p;k++)
        {
            if( fabs(ks[k])> stat[k])
                stat[k] = fabs(ks[k]);
        }
        
        
    } /*fin du sup */
    
    free(var); free(cte); free(ks);
    for(j=0; j<=n; j++)
        free(MM[j]);
    free(MM);    
    
    
} 




void bds_rank_lebaron_sup_prod(int *rangs, int n, int p, double *stat)  /*** MN_star  ***/
/*  =====  The statistics $\tilde M_n^\star$ in our JASA 2006 Paper  ====*/

{
    
    
    double *NN, *var, eps, *cte, *ks;
    int iep, k;
    
    cte = (double *)calloc(p+1,sizeof(double));
    var = (double *)calloc(p+1,sizeof(double));
    ks  = (double *)calloc(p+1,sizeof(double));
    
    for(k=2;k<=p;k++)
    {
        var[k] = 1.0;
        stat[k] = 0.0;
    }   
    
    
    NN = (double *)calloc(n+1,sizeof(double)); 
    
    
    for(iep=1;iep<=n;iep++)
    {
        eps = 1.0*iep/((double) n+1);
        puiss_BDS_star(p, n,cte, eps);
        vec_BDS(n, NN, eps);
        
        bds_rank_lebaron_prod(rangs, n, p, ks, NN, cte, var);
        
        for(k=2;k<=p;k++)
        {
            if( fabs(ks[k])> stat[k])
                stat[k] = fabs(ks[k]);
        }
        
        
    } /*fin du sup */
    
    
    free(var); free(cte); free(ks);
    free(NN); 
} 

/********** In_star ********************/


void bds_rank_prod_int(int *rangs, int n, int p, double *stat)   
/*  =====  The statistics $\tilde I_n^\star$ in our JASA 2006 Paper  ====*/

{
    
    int i,j,k;
    
    double m,a, *sum1,*xx;
    
    sum1 = (double *)calloc(p+1,sizeof(double));
    xx = (double *)calloc(p,sizeof(double));
    for(k=2;k<=p;k++)
    {
        sum1[k] = 0.0; 
    }
    
    for(i=0; i<n ;i++)
    {  
        for(j=0;j<p;j++){xx[j]=rangs[i+j]/(n+1.0); }
        for(k=2;k<=p;k++)
        {
            a  =  bds_prod_integral_comp(xx,k);
            /* printf("\n a= %f k=%d", a,k); */
            sum1[k] += a;
        }
    }
    
    
    for(k=2;k<=p;k++){
        m = (double) n;
        stat[k] =  2.0*sqrt(m)*(sum1[k]/m -Gtilde_n(k,n,Bias_Correction));
    }
    free(sum1); free(xx);
} 





