 /*********************************************************************/
 /* programme: series.c                                               */
 /*                                                                   */
 /* utilit: Programme qui gnre,  partir de la dernire occurence  */
 /*          d'une srie chronologique, une valeur selon diffrentes  */
 /*          lois de probabilits                                     */
 /*                                                                   */
 /* fonction utilise: kiss.c ( gnration de nombres distribus      */  
 /*                                  uniformment sur [0,1]     )     */
 /*  Lois:                                                            */
 /*       1-  Modle autorgressif avec erreurs de loi Normale        */
 /*       2-  Modle autorgressif avec erreurs de loi logistique     */                                         
 /*       3-  Modle autorgressif avec erreurs de loi de Laplace     */                                              
 /*       4-  Modle autorgressif avec erreurs de loi de Cauchy      */
 /*       5-  Modle MA                                               */ 
 /*       6-  Modle NLMA                                             */  
 /*       7-  Modle GARCH                                            */  
 /*       8-  Modle ARCH                                             */
 /*       9-  Tent map                                               */
 /*       10-  Modle MART                                             */
 /*       11-  Feigenbaum map                                          */
 /*       12-  Modle threshold de Tong et Lim                        */   
 /*      13-  Copule de Clayton                                       */
 /*      14-  Copule de Frank                                         */
 /*      15-  Copule de Ali-Mikhail                                   */
 /*      16-  Copule de Farlie-Gumbel-Morgenstern                     */
 /*                                                                   */
 /* par: Jean-Franois Quessy, le 24 fvrier 2000                     */
 /* modifi par  Christian Genest et Bruno Remillard, le 19 novembre 2005 */
 /*********************************************************************/






#include <stdio.h>
#include <stdlib.h>  
#include <math.h>  
#include <malloc.h> 
#include <sys/types.h>
#include <time.h>
#include "kiss.c"

#define pi 3.142592654

   double h(double u)
   { 
      return sqrt(fabs(u));
   }

   double tent(double u)
   
   {
      if( u < 0.5)
         return 1.99999999999*u;    
      else 
         return 1.99999999999*(1.0 - u);
   }

   double feig(double u)
   {
      return 4.0*u*(1.0-u);
   }

 /*  Gaussian variate */
   double rnorm() 
   {
      double u1, u2, R, S;
      u1 = kiss(); 
      u2 = kiss();
      R = pow((double)-2*log(u1),(double)0.5);
      S = cos(2*pi*u2); 
      return R*S;
   }

/* Logistic variate */ 

   double rlogis()  
   {
      double u1;
      u1 = kiss(); 
   
      return log(u1/(1.0-u1));
   }

   double rlap()   /* Laplace variate */
   {
      double u1, u2;
   
      u1 = kiss(); 
      u2 = kiss();
   
      if (u1<=0.5)
      {
         return log(2*u2);
      }
      else
      {
         return -log(2-2*u2);
      }
   }


   double rcau()
   {
      double u1;
   
      u1 = kiss();
   
      return tan(pi*(u1-0.5));
   }

   double rbeta()
   {
      return 0.5*(1.0 -cos(pi*kiss()));
   }
   void laws(double *x, int n, int loi, double theta)
   {
      double u, t, u1, u2, u3, racine, temp;
      double a, b, c1, c2, c3, c4, t1, t2, t3, t4, t5, t6, t7, t8, t9;
      double first; 
      int k,i, j;
   
   
   
      if (loi == 1)  /** AR(1) Model with Gaussian errors **/
      {
         x[0] = sqrt(1.0-theta*theta)*rnorm();
         for(k=1;k<n;k++)
            x[k] = theta*x[k-1] + rnorm();
      }
      
      
      
      else if (loi == 2)  /** Modele AR avec erreurs de loi Logistique **/
      {
         temp = 0.0;
         for(i=0;i<20;i++)
            temp = theta*temp + rlogis();
      
         x[0] = temp;
         for(k=1;k<n;k++)
            x[k] = theta*x[k-1] + rlogis();
      }
      
      
      /** Modele AR avec erreurs de loi de Laplace **/
      else if (loi == 3)
      {
      
         temp = 0.0;
         for(i=0;i<20;i++)
            temp = theta*temp + rlap();
      
         x[0] = temp;
         for(k=1;k<n;k++)
            x[k] = theta*x[k-1] + rlap();
      }
      
      
      /** Modele AR avec erreurs de loi de Cauchy **/
      else if (loi == 4)
      {
         x[0] = rcau()/(1.0 - fabs(theta));
      
         for(k=1;k<n;k++)
            x[k] = theta*x[k-1] + rcau();
      }
      
      /** Modele MA **/
      
      else if (loi == 5)
      {
         u1 = rnorm();
      
      
         for(k=0;k<n;k++)
         { 
            u2 = rnorm();
            x[k] = u2 - theta*u1;
            u1 = u2;
         }
      
      }
      
      
        /** Modele NLMA **/
      
      else if (loi == 6)
      {
         u1 = rnorm();
         u2 = rnorm();
      
         for(k=0;k<n;k++)
         { 
            u3 = rnorm();
            x[k] = u3+ theta*u1*u2;
            u1 = u2;
            u2 = u3;
         
         }
      
      }
       
      
      /** Modele GARCH **/
      else if (loi == 7)
      {
         x[0]=rnorm();
         temp=1.0;
         for(k=1;k<20;k++)
         {
            u2 = rnorm();
            temp=sqrt(1.0 +theta*temp*temp+ 2*theta*x[k-1]*x[k-1]);
            x[k]=u2*temp;
         }
         x[0]=x[19];
         for(k=1;k<n;k++)
         {
            u2 = rnorm();
            temp=sqrt(1.0 +theta*temp*temp+ 2*theta*x[k-1]*x[k-1]);
            x[k]=u2*temp;
         }
      
      }
      
      
      /** Modele ARCH **/
      else if (loi == 8)
      {
         temp = 0.0;
         for(i=0;i<20;i++)
            temp = rnorm()*sqrt(1.0+ theta*temp*temp);
      
         x[0] = temp;
         for(k=1;k<n;k++)
            x[k] = rnorm()*sqrt(1.0+ theta*x[k-1]*x[k-1]);
      
      }
      
      
      /** Tent Map **/   
      
      else if (loi == 9)
      {
      
      
         x[0] = kiss();
      
      
         for(k=1;k<n;k++)
         { 
            u1 = kiss(); 
         
            if ( u1 < 4.0*theta)
               x[k] = tent(x[k-1]);
            else
               x[k] = kiss();
         }
      }
      
      /** Modele Mart **/
      else if (loi == 10) 
      {  
         temp = 0.0;
      
         for(i=0;i<20;i++)
            temp = rnorm()*(1.0+ sqrt(theta)*temp );   
      
         x[0] = temp;  
      
         for(k=1;k<n;k++) 
            x[k] = rnorm()*(1.0 + sqrt(theta)*x[k-1] ); 
      
      
      } 
      
      /** Logistic Map **/   
      else if (loi == 11)  
      {
         x[0] = rbeta();
      
         for(k=1;k<n;k++)
         { 
            u1 = kiss(); 
         
            if ( u1 < 4.0*theta)
               x[k] = feig(x[k-1]);
            else
               x[k] = rbeta();
         }
      } 
      
      
      
      /** Threshold model of Tong and Lim (1980) **/ 
      else if (loi == 12)  
      {    
         temp = kiss();
      
         for(i=0;i<20;i++)    
         {
            u1 = kiss();
            if (temp > 0.5)
               temp =  u1 - theta*temp;
            else
               temp = u1+ theta*temp;
         }
         x[0] = temp;
      
         for(k=1;k<n;k++)    
         {
            u1 = kiss();
            if (x[k-1] > 0.5)
               x[k] =  u1 - theta*x[k-1];
            else
               x[k] = u1+ theta*x[k-1];
         }
      
      } 
      
      
      
      /** Loi de Clayton **/
      else if (loi == 13)
      {
      
         if (theta == 0.0) 
         {
            for(k=0;k<n;k++) 
               x[k] = kiss();
         }
         else
         {
            c1 = (double)theta/(1.0+theta); 
            c2 = -1.0/theta;
         
         
            x[0] = kiss();
            for(k=1;k<n;k++)
            {
               u1 = kiss();
               a = pow( (double)x[k-1], -(double)theta );
               b = pow( (double)u1, -(double)c1 );
               x[k] = pow( (double) 1.0+ a*(b-1.0), (double)c2 );
            }
         }
      }         
      
      
      
      /** Loi  Frank **/
      else if (loi == 14)
      {
         if (theta == 0.0) 
         {
            for(k=0;k<n;k++) 
               x[k] = kiss();
         }
         else
         {
            x[0] = kiss();
            for(k=1;k<n;k++)
            {
               u = x[k-1]; 
               t = kiss();
            
               t1 = t*exp(-1.0*theta) + (1.0-t)*exp(-1.0*theta*u);
               t2 = t + (1.0-t)*exp(-1.0*theta*u);
            
               x[k] = log(t2/t1)/theta;
            }
         }
      }
      
      
      /** Loi de de Ali-Mikhail-Haq **/
      else if (loi == 15)
      {
         if (theta == 0.0) 
         {
            for(k=0;k<n;k++) 
               x[k] = kiss();
         }
         else
         {
            x[0] = kiss();
            for(k=1;k<n;k++)
            {
               u = x[k-1];
               t = kiss();
            
               t1 = t*theta;
               t2 = pow(theta, 2.0);
               t3 = t2*t;
               t4 = u*u;
               t5 = t4*t3;
               t6 = u*t1;
               t7 = u*t3;
               t8 = 4.0*t7;
               t9 = sqrt(1.0-2.0*theta+t2+4.0*t5-t8+4.0*t6);
            
               x[k] = (1/(t3-2.0*t7-theta+t5)*
                      (-theta-2.0*t1+1.0+2.0*t3+2.0*t5+2.0*t6-t8-t9)/2.0); 
            }
         }
      }
      
      
      /** Loi de Farlie-Gumbel-Morgenstern **/
      else if (loi == 16)
      {  
         x[0] = kiss();
         for(k=1;k<n;k++)
         {
            t = kiss();
            b = theta*(1.0-2.0*x[k-1]);
            racine = sqrt((1.0 + b)*(1.0 + b)-4.0*t*b );
            x[k] = 2.0*t/(1.0+b+racine);
         }
      } 
   }
